#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
from django.contrib import admin

from django_quotas.config import DjangoQuotasConfig as cfg
from django_quotas.defaults.models import DefaultQuotaModel


@admin.register(DefaultQuotaModel)
class DefaultQuotaAdmin(cfg.base_admin_cls()):  # type: ignore[misc]
    pass



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
from functools import cached_property
from typing import TYPE_CHECKING, Final

from django.conf import settings

from django_quotas.utils import get_model_by_name

if TYPE_CHECKING:
    from django_quotas.backends.db.models import QuotaUsageModel


__all__ = ["__DjangoQuotasDbConfig", "DjangoQuotasDbConfig"]


class __DjangoQuotasDbConfig:
    """Configuration accessor for DB-backed quotas implementation."""

    SETTINGS_PREFIX: Final[str] = "DJANGO_QUOTAS"

    @cached_property
    def QUOTA_USAGE_MODEL(self) -> str:
        """Return the full model name for the quota usage model.

        @return: Model name string.
        """
        return getattr(settings, f"{self.SETTINGS_PREFIX}_DB_USAGE_MODEL_NAME", "django_quotas_db.QuotaUsageModel")

    @cached_property
    def quota_usage_cls(self) -> type["QuotaUsageModel"]:
        """Return the quota usage model class.

        @return: Quota usage model class.
        """
        return get_model_by_name(self.QUOTA_USAGE_MODEL)  # type: ignore


DjangoQuotasDbConfig = __DjangoQuotasDbConfig()



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
import abc
from collections import defaultdict
import uuid

from asgiref.sync import sync_to_async

from django_quotas.base.dto import Quota, QuotaBucket, QuotaStats, QuotaUseForBucket, ValuePerBucket

__all__ = [
    "QuotaExceededError",
    "QuotaService",
]


class QuotaExceededError(Exception):
    """Exception raised when a quota is exceeded for an account and feature(s).

    :param account_name: The account UUID.
    :param exceeded_features: Mapping of feature names to exceeded bucket stats.
    """

    def __init__(self, account_name: uuid.UUID, exceeded_features: dict[str, list[QuotaUseForBucket]]):
        self.account_id = account_name
        self._exceeded_features_stats = exceeded_features
        super().__init__(self.__generate_detailed_message())

    def get_exceeded_features(self) -> set[str]:
        """Get the set of feature names for which quotas are exceeded.

        :return: Set of feature names.
        """
        return set(self._exceeded_features_stats.keys())

    def get_stats_per_feature(self) -> dict[str, list[QuotaUseForBucket]]:
        """Get the exceeded stats per feature.

        :return: Mapping of feature names to list of QuotaUseForBucket.
        """
        return self._exceeded_features_stats

    def __generate_detailed_message(self) -> str:
        msg = self.__generate_msg() + "\nDetails:"
        for feature, stats in self._exceeded_features_stats.items():
            msg += f"\n{feature}: "
            buckets = []
            for stat in stats:
                buckets.append(f"{stat.bucket_name}({stat.current_usage}/{stat.limit})")
            msg += ", ".join(buckets)
        return msg

    def __generate_msg(self) -> str:
        return f"Quota exceeded for account {self.account_id}. Exceeded features: {self.get_exceeded_features()}"


class QuotaService(metaclass=abc.ABCMeta):
    """Abstract base class for quota service implementations."""

    def ensure_quota_or_raise(
        self, account_id: uuid.UUID, feature_name: str | set[str], potential_increase: int = 1
    ) -> None:
        """
        Ensure that the quota for the given account and feature is not exceeded.

        If the quota is exceeded, raise a QuotaExceededError.
        :param account_id: The account ID.
        :param feature_name: The feature name or set of feature names.
        :param potential_increase: The potential increase in usage.\
            Current utilization + potential increase must be less than the quota.
        :raise QuotaExceededError: If the quota is exceeded.
        """
        feature_names: set[str] = {feature_name} if isinstance(feature_name, str) else feature_name
        utilization = self.get_quotas_utilization(account_id, feature_name)
        exceeded_features: dict[str, list[QuotaUseForBucket]] = defaultdict(list)
        for feature in feature_names:
            if feature not in utilization.feature_stats:
                continue  # this means we have no quota for this specific feature
            status = utilization.feature_stats[feature]
            quota_buckets = [
                (QuotaBucket.HOURLY, status.usage.hourly or 0, status.limits.hourly),
                (QuotaBucket.DAILY, status.usage.daily or 0, status.limits.daily),
                (QuotaBucket.MONTHLY, status.usage.monthly or 0, status.limits.monthly),
                (QuotaBucket.TOTAL, status.usage.total or 0, status.limits.total),
            ]

            for bucket_name, current_usage, limit in quota_buckets:
                if limit is not None and current_usage + potential_increase > limit:
                    exceeded_features[feature].append(
                        QuotaUseForBucket(
                            bucket_name=bucket_name,
                            current_usage=current_usage,
                            limit=limit,
                            quota_id=status.quota_id,
                        )
                    )
        if exceeded_features:
            raise QuotaExceededError(account_id, exceeded_features)

    async def aensure_quota_or_raise(
        self, account_id: uuid.UUID, feature_name: str | set[str], potential_increase: int = 1
    ) -> None:
        """
        Ensure that the quota for the given account and feature is not exceeded.

        If the quota is exceeded, raise a QuotaExceededError.
        :param account_id: The account ID.
        :param feature_name: The feature name.
        :param potential_increase: The potential increase in usage.\
            Current utilization + potential_increase must be less than the quota.
        """
        return await sync_to_async(self.ensure_quota_or_raise)(account_id, feature_name, potential_increase)

    @abc.abstractmethod
    def register_usage(self, account_id: uuid.UUID, feature_name: str, increment: int = 1) -> None:
        """
        Register usage for the given account and feature.

        :param account_id: The account ID.
        :param feature_name: The feature name.
        :param increment: The value to be added to quota usage.
        """
        pass

    @abc.abstractmethod
    async def aregister_usage(self, account_id: uuid.UUID, feature_name: str | set[str], increment: int = 1) -> None:
        """
        Register usage for the given account and feature.

        :param account_id: The account ID.
        :param feature_name: The feature name.
        :param increment: The value to be added to quota usage.
        """
        pass

    @abc.abstractmethod
    def get_quotas_utilization(self, account_id: uuid.UUID, feature_name: str | set[str] | None) -> QuotaStats:
        """
        Get the quota utilization for the given account and features.

        :param account_id: The account ID.
        :param feature_name: Name of the feature or a list of feature names. None means all features having quotas.
        """
        pass

    @abc.abstractmethod
    async def aget_quotas_utilization(self, account_id: uuid.UUID, feature_name: str | set[str] | None) -> QuotaStats:
        """
        Get the quota utilization for the given account and features.

        :param account_id: The account ID.
        :param feature_name: Name of the feature or a list of feature names. None means all features having quotas.
        """
        pass

    @abc.abstractmethod
    def set_quota(
        self, account_id: uuid.UUID, feature_name: str, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        """
        Set the quota for the given account and feature.

        :param account_id: The account ID.
        :param feature_name: The feature name.
        :param limits: Limits for each bucket.
        :param owner_tag: Optional owner tag.
        :return: The created or updated quota.
        """
        pass

    @abc.abstractmethod
    async def aset_quota(
        self, account_id: uuid.UUID, feature_name: str, limits: ValuePerBucket, owner_tag: str | None = None
    ) -> Quota:
        """
        Set the quota for the given account and feature.

        :param account_id: The account ID.
        :param feature_name: The feature name.
        :param limits: Limits for each bucket.
        :param owner_tag: Optional owner tag.
        :return: The created or updated quota.
        """
        pass



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
"""Django quotas management library package root.

@package django_quotas
"""

__all__ = []

#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
from functools import cached_property
from typing import TYPE_CHECKING, Final, cast

from django.conf import settings
from django.db import models

from django_quotas.utils import get_class_by_name, get_model_by_name

if TYPE_CHECKING:
    from django.contrib.admin import ModelAdmin

    from django_quotas.models import BaseQuotaModel


__all__ = ["DjangoQuotasConfig"]


class __DjangoQuotasConfig:
    """Configuration accessor for django_quotas settings and related models."""

    SETTINGS_PREFIX: Final[str] = "DJANGO_QUOTAS"

    @cached_property
    def TABLE_PREFIX(self) -> str:
        """Return the table prefix for quota tables.

        :return: Table prefix string.
        """
        return getattr(settings, f"{self.SETTINGS_PREFIX}_TABLE_PREFIX", "django_quotas")

    @cached_property
    def QUOTA_MODEL(self) -> str:
        """Get the full model name for the quota model.

        :return: Model name string.
        """
        return getattr(settings, f"{self.SETTINGS_PREFIX}_QUOTA_MODEL_NAME", "django_quotas.defaults.QuotaModel")

    @cached_property
    def QUOTA_RELATED_ACCOUNT_MODEL(self) -> str:
        """Get the related account model name for quotas.

        :return: Related account model name string.
        """
        return getattr(settings, f"{self.SETTINGS_PREFIX}_QUOTA_RELATED_ACCOUNT_MODEL_NAME", "auth.User")

    @cached_property
    def BASE_ADMIN_CLASS(self) -> str:
        """Get the base admin class for quota admin interfaces.

        :return: Base admin class name string
        """
        return getattr(settings, f"{self.SETTINGS_PREFIX}_BASE_ADMIN_CLASS", "django.contrib.admin.ModelAdmin")

    @cached_property
    def quota_cls(self) -> type["BaseQuotaModel"]:
        """Get the quota model class.

        :return: Quota model class.
        """
        return get_model_by_name(self.QUOTA_MODEL)  # type: ignore

    @cached_property
    def quota_related_account_cls(self) -> type[models.Model]:
        """Get the related account model class for quotas.

        :return: Related account model class.
        """
        return get_model_by_name(self.QUOTA_RELATED_ACCOUNT_MODEL)

    def base_admin_cls(self) -> type["ModelAdmin"]:
        """Get the base admin class for quota admin interfaces.

        :return: Base admin class.
        """
        return cast(type["ModelAdmin"], get_class_by_name(self.BASE_ADMIN_CLASS))


DjangoQuotasConfig = __DjangoQuotasConfig()



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
"""Database-backed implementation subpackage for django_quotas.

@package django_quotas.backend.db
"""

__all__ = []

#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
import abc
import dataclasses
import datetime
from enum import StrEnum
import uuid

__all__ = [
    "QuotaBucket",
    "ValuePerBucket",
    "QuotaStatus",
    "QuotaStats",
    "QuotaUseForBucket",
    "Quota",
    "QuotaUsage",
]


class QuotaBucket(StrEnum):
    """Enumeration of supported quota buckets."""

    HOURLY = "hourly"
    DAILY = "daily"
    MONTHLY = "monthly"
    TOTAL = "total"


@dataclasses.dataclass(kw_only=True)
class ValuePerBucket:
    """Hold quota values for each bucket type.

    :param hourly: Hourly quota value.
    :param daily: Daily quota value.
    :param monthly: Monthly quota value.
    :param total: Total quota value.
    """

    hourly: int | None = None
    daily: int | None = None
    monthly: int | None = None
    total: int | None = None


@dataclasses.dataclass(kw_only=True)
class QuotaStatus:
    """Represent the status of a quota, including limits and usage.

    :param quota_id: Unique identifier for the quota.
    :param limits: Quota limits per bucket.
    :param usage: Current usage per bucket.
    """

    quota_id: uuid.UUID
    limits: ValuePerBucket
    usage: ValuePerBucket

    def is_exceeded(self) -> bool:
        """Check if any quota limit is exceeded.

        @return: True if any quota is exceeded, False otherwise.
        """
        if self.limits.daily is not None and self.usage.daily is not None and self.usage.daily >= self.limits.daily:
            return True
        if (
            self.limits.monthly is not None
            and self.usage.monthly is not None
            and self.usage.monthly >= self.limits.monthly
        ):
            return True
        if self.limits.total is not None and self.usage.total is not None and self.usage.total >= self.limits.total:
            return True
        return False


@dataclasses.dataclass(kw_only=True)
class QuotaStats:
    """Aggregate quota status for an account and its features.

    :param account_id: Account identifier.
    :param feature_stats: Mapping of feature names to their quota status.
    """

    account_id: uuid.UUID
    feature_stats: dict[str, QuotaStatus]

    def has_exceeded_quotas(self) -> bool:
        """Check if any feature has exceeded quotas.

        @return: True if any feature has exceeded its quota, False otherwise.
        """
        return any(status.is_exceeded() for status in self.feature_stats.values())


@dataclasses.dataclass(kw_only=True)
class QuotaUseForBucket:
    """Usage and limit information for a specific quota bucket.

    :param bucket_name: Name of the quota bucket.
    :param current_usage: Current usage value.
    :param limit: Limit for the bucket.
    :param quota_id: Associated quota identifier.
    """

    bucket_name: QuotaBucket
    current_usage: int
    limit: int | None
    quota_id: uuid.UUID

    @property
    def overuse(self) -> int | None:
        """Return the amount by which the usage exceeds the limit, or None if no limit.

        @return: Overuse amount or None.
        """
        if self.limit is None:
            return None
        return self.current_usage - self.limit


class Quota(metaclass=abc.ABCMeta):
    """Abstract base class for quota definitions."""

    @property
    @abc.abstractmethod
    def id(self) -> uuid.UUID:
        """Return the unique identifier for the quota.

        @return: Quota UUID.
        """
        pass

    @property
    @abc.abstractmethod
    def account_id(self) -> uuid.UUID:
        """Return the account identifier associated with the quota.

        @return: Account UUID.
        """
        pass

    @property
    @abc.abstractmethod
    def feature_name(self) -> str:
        """Return the feature name associated with the quota.

        @return: Feature name string.
        """
        pass

    @property
    @abc.abstractmethod
    def get_limits(self) -> ValuePerBucket:
        """Return the quota limits for each bucket.

        @return: ValuePerBucket instance.
        """
        pass


class QuotaUsage(metaclass=abc.ABCMeta):
    """Abstract base class for quota usage records."""

    @property
    @abc.abstractmethod
    def account_id(self) -> uuid.UUID:
        """Return the account identifier for the usage record.

        @return: Account UUID.
        """
        pass

    @property
    @abc.abstractmethod
    def feature_name(self) -> str:
        """Return the feature name for the usage record.

        @return: Feature name string.
        """
        pass

    @property
    @abc.abstractmethod
    def point_in_time(self) -> datetime.datetime:
        """Return the timestamp of the usage record.

        @return: Datetime of usage.
        """
        pass

    @property
    @abc.abstractmethod
    def usage_count(self) -> int:
        """Return the usage count for the record.

        @return: Usage count as integer.
        """
        pass



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
import uuid

from django.db import models
from django.utils.translation import gettext_lazy as gt

from django_quotas.config import DjangoQuotasConfig as cfg


class QuotaUsageModel(models.Model):
    """Model for tracking quota usage per account, feature, and time point."""

    class Meta:
        db_table = f"{cfg.TABLE_PREFIX}_quota_usage"
        verbose_name = gt("Quota Usage")
        unique_together = ("account", "feature_name", "point_in_time")
        indexes = (
            models.Index(fields=["account", "feature_name", "point_in_time"]),
            models.Index(fields=["account", "point_in_time"]),
        )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    account = models.ForeignKey(  # type: ignore[call-arg]
        cfg.QUOTA_RELATED_ACCOUNT_MODEL, null=False, blank=False, on_delete=models.CASCADE, swappable=True
    )
    feature_name = models.CharField(max_length=300, null=False, blank=False)
    point_in_time = models.DateTimeField(null=False, blank=False)
    usage_count = models.IntegerField(null=False, blank=False, default=0)

    @property
    def account_id(self) -> uuid.UUID:  # type: ignore[no-untyped-def]
        """Return the unique identifier for the associated account.

        :return: Account UUID.
        """
        return self.account.pk



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#
from django.contrib import admin
from django.http import HttpRequest

from django_quotas.backends.db.models import QuotaUsageModel
from django_quotas.config import DjangoQuotasConfig as cfg


@admin.register(QuotaUsageModel)
class QuotaUsageAdmin(cfg.base_admin_cls()):  # type: ignore[misc]
    def has_add_permission(self, request: HttpRequest) -> bool:
        return False



#
#  Copyright 2025 by Dmitry Berezovsky, MIT License
#

import uuid

from django.db import migrations, models
import django.db.models.deletion

from django_quotas.config import DjangoQuotasConfig as cfg


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(cfg.QUOTA_RELATED_ACCOUNT_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="QuotaUsageModel",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("feature_name", models.CharField(max_length=300)),
                ("point_in_time", models.DateTimeField()),
                ("usage_count", models.IntegerField(default=0)),
                (
                    "account",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=cfg.QUOTA_RELATED_ACCOUNT_MODEL),
                ),
            ],
            options={
                "verbose_name": "Quota Usage",
                "db_table": "django_quotas_quota_usage",
                "indexes": [
                    models.Index(
                        fields=["account", "feature_name", "point_in_time"], name="django_quot_account_b2afd6_idx"
                    ),
                    models.Index(fields=["account", "point_in_time"], name="django_quot_account_6a9881_idx"),
                ],
                "unique_together": {("account", "feature_name", "point_in_time")},
            },
        ),
    ]
